"use strict";

var _toArray = function (arr) {
  return Array.isArray(arr) ? arr : Array.from(arr);
};

var _prototypeProperties = function (child, staticProps, instanceProps) {
  if (staticProps) Object.defineProperties(child, staticProps);
  if (instanceProps) Object.defineProperties(child.prototype, instanceProps);
};

var _get = function get(object, property, receiver) {
  var desc = Object.getOwnPropertyDescriptor(object, property);

  if (desc === undefined) {
    var parent = Object.getPrototypeOf(object);

    if (parent === null) {
      return undefined;
    } else {
      return get(parent, property, receiver);
    }
  } else if ("value" in desc && desc.writable) {
    return desc.value;
  } else {
    var getter = desc.get;
    if (getter === undefined) {
      return undefined;
    }
    return getter.call(receiver);
  }
};

var _inherits = function (child, parent) {
  if (typeof parent !== "function" && parent !== null) {
    throw new TypeError("Super expression must either be null or a function, not " + typeof parent);
  }
  child.prototype = Object.create(parent && parent.prototype, {
    constructor: {
      value: child,
      enumerable: false,
      writable: true,
      configurable: true
    }
  });
  if (parent) child.__proto__ = parent;
};

var _interopRequireWildcard = function (obj) {
  return obj && obj.constructor === Object ? obj : {
    "default": obj
  };
};

var path = require("path");
var cp = require("child_process");
var semver = require("semver");
var semverRegex = require("semver-regex");
var dargs = require("dargs");
var through = require("through2");
var concat = require("concat-stream");
var utils = _interopRequireWildcard(require("./utils"));

var which = utils.denodeify(require("which"));

var BIN = "sass-convert";
var MINVERS = ">=3.4.5";

/**
 * Execute command in a child process.
 *
 * @see {@link http://nodejs.org/api/child_process.html}
 * @param {String} command
 * @param {Array} args
 * @return {Promise}
 */
function exec(command) {
  var args = [];

  for (var _key = 1; _key < arguments.length; _key++) {
    args[_key - 1] = arguments[_key];
  }

  var deferred = utils.defer();
  var childProcess = undefined;

  args.push(function (err, stdout, stderr) {
    if (err) {
      Object.assign(err, {
        message: "" + err.message + " `" + command + "` exited with error code " + err.code,
        stdout: stdout,
        stderr: stderr });

      deferred.reject(err);
    } else {
      deferred.resolve({
        childProcess: childProcess,
        stdout: stdout,
        stderr: stderr });
    }
  });

  childProcess = cp.exec.apply(cp, [command].concat(_toArray(args)));

  // process.nextTick(() => {
  //   deferred.notify(childProcess);
  // });

  return deferred.promise;
}

/**
 * Custom error for binary check.
 *
 * @param {String} message
 */
var BinaryError = (function () {
  var _Error = Error;
  var BinaryError = function BinaryError(message) {
    _get(Object.getPrototypeOf(BinaryError.prototype), "constructor", this).call(this, message);
    // http://bit.ly/1yMzARU
    this.message = message || "Could not find any executable for \"" + BIN + "\". Operation Aborted.";
  };

  _inherits(BinaryError, _Error);

  _prototypeProperties(BinaryError, null, {
    name: {
      get: function () {
        return "BinaryError";
      },
      enumerable: true
    }
  });

  return BinaryError;
})();

/**
 * Custom error for version check.
 *
 * @param {String} message
 */
var VersionError = (function () {
  var _Error2 = Error;
  var VersionError = function VersionError(message) {
    _get(Object.getPrototypeOf(VersionError.prototype), "constructor", this).call(this, message);
    // http://bit.ly/1yMzARU
    this.message = message || "Invalid \"" + BIN + "\" version, must be " + MINVERS;
  };

  _inherits(VersionError, _Error2);

  _prototypeProperties(VersionError, null, {
    name: {
      get: function () {
        return "VersionError";
      },
      enumerable: true
    }
  });

  return VersionError;
})();

/**
 * Check whether passed binary (Gem) is in $PATH,
 * and check its version.
 *
 * @param {String} bin
 * @return {Promise}
 */
function checkBinary(bin) {
  /**
   * Check for binary min-version.
   *
   * @param {String} str
   * @return {Boolean}
   */
  function checkVersion(str) {
    var version = str.match(semverRegex())[0];
    return semver.satisfies(version, MINVERS);
  }

  /**
   * Check for global binary and version.
   *
   * @return {Promise}
   */
  function checkGlobal() {
    return which(bin).then(function () {
      return exec("" + bin + " -v");
    }, function () {
      throw new BinaryError();
    }).then(function (res) {
      if (!checkVersion(res.stdout)) {
        throw new VersionError();
      }
    });
  }

  /**
   * Check for bundled binary and version.
   *
   * @return {Promise}
   */
  function checkBundle() {
    return which("bundle").then(function () {
      return exec("bundle exec " + bin + " -v");
    }).then(function (res) {
      if (!checkVersion(res.stdout)) {
        throw new VersionError();
      }
      return { bundler: true };
    }, function (err) {
      throw new BinaryError();
    });
  }

  return checkBundle()["catch"](checkGlobal);
}

/**
 * Run binary checks only once.
 *
 * @return {Promise}
 */
function checkBinaryCache() {
  if (!("promise" in checkBinaryCache)) {
    checkBinaryCache.promise = checkBinary(BIN);
  }

  return checkBinaryCache.promise;
}

var Converter = function Converter(options) {
  this.options = options || {};
  this.bundler = false;
};

/**
 * Format the `sass-convert` command string.
 *
 * @return {Object}
 */
Converter.prototype.command = function () {
  var cmd = "" + (this.bundler ? "bundle exec " : "") + "sass-convert";

  // Add required args.
  var options = Object.assign(this.options, {
    stdin: true,
    "no-cache": true });

  // Filter unwanted or erroneous args.
  var allowed = ["from", "to", "dasherize", "indent", "old", "default-encoding", "unix-newlines", "stdin", "no-cache"];

  var args = dargs(options, [], allowed);

  return { cmd: cmd, args: args };
};

/**
 * Returns whether file is eligible for convertion.
 *
 * @param {Object} file - Vinyl file Object
 * @return {Boolean}
 */
Converter.prototype.isTransformable = function (file) {
  var ext = path.extname(file.path);
  return (file.isBuffer() || file.isStream()) && ext.endsWith(this.options.from);
};

/**
 * Convert file Buffer through `sass-convert` binary.
 *
 * @param {Buffer} file - Vinyl file
 * @param {String} enc - encoding
 */
Converter.prototype.convertFile = function (file, enc) {
  var deferred = utils.defer();
  var _ref = this.cmdCache || (this.cmdCache = this.command());
  var cmd = _ref.cmd;
  var args = _ref.args;
  var child = cp.spawn(cmd, args);
  var result = {};

  file.pipe(child.stdin);

  child.stdout.pipe(concat(function (data) {
    result.stdout = data;
  }));

  child.stderr.pipe(concat(function (data) {
    result.stderr = data;
  }));

  child.on("error", deferred.reject);

  child.on("close", function (code, signal) {
    if (code !== 0) {
      var message = "`" + cmd + " " + args.join(" ") + "` failed with code " + code;
      var stderr = Buffer.isBuffer(result.stderr) ? result.stderr.toString() : "";

      deferred.reject({ code: code, message: message, stderr: stderr });
    } else {
      deferred.resolve(result);
    }
  });

  return deferred.promise;
};

/**
 * Returns a transform stream meant to be piped to a stream
 * of Sass, SCSS or CSS files. Apply convertion if matches.
 *
 * @return {Object} - Stream
 */
Converter.prototype.stream = function () {
  var self = this;

  return through.obj(function (file, enc, cb) {
    var _stream = this;

    if (!self.isTransformable(file)) {
      // Pass-through.
      return cb(null, file);
    }

    // Matches `from`, let's convert it.
    checkBinaryCache().then(function (res) {
      self.bundler = res ? res.bundler : false;

      return self.convertFile(file, enc);
    }).then(function (result) {
      file.contents = new Buffer(result.stdout);

      cb(null, file);
    }, function (err) {
      _stream.emit("error", err);

      if (self.options.force) {
        // Continue stream chain,
        // but don't pass unconverted chunks.
        return cb();
      }

      // Stop stream chain.
      _stream.destroy();
      _stream.emit("end");
    });
  });
};

module.exports = function (options) {
  return new Converter(options).stream();
};

// Expose custom error constructors.
module.exports.BinaryError = BinaryError;
module.exports.VersionError = VersionError;